#!/bin/env python
import os

from setuptools import setup, find_packages
import distutils
from distutils.core import Extension

# Makefile:
# CFLAGS = -Wall -Wextra
# CFLAGS_RELEASE = $(CFLAGS) -O2 -DNDEBUG
# CFLAGS_DEBUG = $(CFLAGS) -g3 -O0 -DDEBUG=1
extra_compile_args = [
    '-Wall',
    '-Wextra',
    '-Werror',
    '-Wfatal-errors',
    # Some internal Python library code does not like this.
    '-Wno-c++11-compat-deprecated-writable-strings',
    '-std=c++17',
    '-Isrc/cpp',
    # We implement mutex with Python's thread locking so we don't want the
    # overhead of C++'s thread locking as well.
    '-USKIPLIST_THREAD_SUPPORT',
    '-Wno-stdlibcxx-not-found',
]

# Horrible hack to support GCC
# To activate this you need to set the USING_GCC environment variable or call
# USING_GCC=1 python setup.py ...
if os.environ.get("USING_GCC"):
    extra_compile_args.append('-Wno-unknown-pragmas')
    extra_compile_args.append('-Wno-cast-function-type')
    extra_compile_args.append('-Wno-missing-field-initializers')

DEBUG = False

if DEBUG:
    extra_compile_args.extend(['-g3', '-O0', '-DDEBUG=1'])
else:
    extra_compile_args.extend(['-g3', '-O2', '-DNDEBUG'])

orderedstructs = Extension(
    "orderedstructs",
    sources=[
        'src/cpy/cOrderedStructs.cpp',
        'src/cpy/OrderedStructs.cpp',
        'src/cpy/cSkipList.cpp',
        'src/cpy/cmpPyObject.cpp',
        'src/cpp/SkipList.cpp',
    ],
    include_dirs=[
        '.',
        # '/Applications/Xcode.app/Contents/Developer/Toolchains/XcodeDefault.xctoolchain/usr/include/c++/v1',
    ],
    library_dirs=[os.getcwd(), ],
    extra_compile_args=extra_compile_args,
    extra_link_args=['-lstdc++'],
    language='c++17',
)

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.md') as history_file:
    history = history_file.read()

requirements = [
]

setup_requirements = [
]

test_requirements = [
    'pytest',
    'hypothesis',
]

setup(
    name='orderedstructs',
    version='0.3.13',
    ext_modules=[orderedstructs, ],
    description="Contains a variety of ordered structures, in particular a SkipList.",
    long_description=readme + '\n\n' + history,
    long_description_content_type = 'text/markdown',
    author="Paul Ross",
    author_email='apaulross@gmail.com',
    url='https://github.com/paulross/skiplist',
#     packages=find_packages('src'),
    license="MIT License",
    keywords=['orderedstructs', 'SkipList'],
    # See: https://pypi.org/classifiers/
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
    test_suite='tests',
    tests_require=test_requirements,
    setup_requires=setup_requirements,
)



__all__ = ["exponentially_weighted_mean"]

import numpy as np

from ._lib import exponentially_weighted as _ew


def exponentially_weighted_mean(x: np.ndarray, alpha: float) -> np.ndarray:
    """Compute the exponentially weighted mean of the input array.

    Args:
        x (np.ndarray): Input array.
        alpha (float): Weight parameter.

    Returns:
        np.ndarray: Array with the exponentially weighted mean.
    """
    return _ew.exponentially_weighted_mean(x, alpha)



__all__ = [
    "expanding_mean",
    "expanding_std",
    "expanding_min",
    "expanding_max",
    "expanding_quantile",
]

from typing import Callable

import numpy as np

from ._lib import expanding as _expanding


def _expanding_stat(x: np.ndarray, stat: str) -> np.ndarray:
    return getattr(_expanding, f"expanding_{stat}")(x)


def _expanding_docstring(*args, **kwargs) -> Callable:
    base_docstring = """Compute the {} of the input array.

    Args:
        x (np.ndarray): Input array.

    Returns:
        np.ndarray: Array with the expanding statistic
    """

    def docstring_decorator(function: Callable):
        function.__doc__ = base_docstring.format(function.__name__)
        return function

    return docstring_decorator(*args, **kwargs)


@_expanding_docstring
def expanding_mean(x: np.ndarray) -> np.ndarray:
    return _expanding_stat(x, "mean")


@_expanding_docstring
def expanding_std(x: np.ndarray) -> np.ndarray:
    return _expanding_stat(x, "std")


@_expanding_docstring
def expanding_min(x: np.ndarray) -> np.ndarray:
    return _expanding_stat(x, "min")


@_expanding_docstring
def expanding_max(x: np.ndarray) -> np.ndarray:
    return _expanding_stat(x, "max")


def expanding_quantile(x: np.ndarray, p: float) -> np.ndarray:
    """Compute the expanding_quantile of the input array.

    Args:
        x (np.ndarray): Input array.
        p (float): Quantile to compute.

    Returns:
        np.ndarray: Array with the expanding statistic
    """
    return _expanding.expanding_quantile(x, p)



import numpy as np

_indptr_dtype = np.int32


def _diffs_to_indptr(diffs: np.ndarray) -> np.ndarray:
    diffs = diffs.astype(_indptr_dtype, copy=False)
    return np.append(
        _indptr_dtype(0),
        diffs.cumsum(dtype=_indptr_dtype),
    )



import numpy as np

from ._lib import seasonal as _seasonal
from .differences import num_seas_diffs


def find_season_length(x: np.ndarray, max_season_length: int) -> int:
    """Find the length of the seasonal period of the time series.
    Returns 0 if no seasonality is found.

    Args:
        x (np.ndarray): Array with the time series.

    Returns:
        int: Season period."""
    period = _seasonal.period(x, max_season_length)
    return num_seas_diffs(x, period, 1) * period



import numpy as np

from ._lib import differences as _diffs


def num_diffs(x: np.ndarray, max_d: int = 1) -> int:
    """Find the optimal number of differences

    Args:
        x (np.ndarray): Array with the time series.
        max_d (int, optional): Maximum number of differences to consider. Defaults to 1.

    Returns:
        int: Optimal number of differences."""
    return _diffs.num_diffs(x, max_d)


def num_seas_diffs(x: np.ndarray, season_length: int, max_d: int = 1) -> int:
    """Find the optimal number of seasonal differences

    Args:
        x (np.ndarray): Array with the time series.
        season_length (int): Length of the seasonal pattern.
        max_d (int, optional): Maximum number of differences to consider. Defaults to 1.

    Returns:
        int: Optimal number of seasonal differences."""
    return _diffs.num_seas_diffs(x, season_length, max_d)


def diff(x: np.ndarray, d: int) -> np.ndarray:
    """Subtract previous values of the series

    Args:
        x (np.ndarray): Array with the time series.
        d (int): Lag to subtract

    Returns:
        np.ndarray: Differenced time series."""
    return _diffs.diff(x, d)



__all__ = [
    "Lag",
    "RollingMean",
    "RollingStd",
    "RollingMin",
    "RollingMax",
    "RollingQuantile",
    "SeasonalRollingMean",
    "SeasonalRollingStd",
    "SeasonalRollingMin",
    "SeasonalRollingMax",
    "SeasonalRollingQuantile",
    "ExpandingMean",
    "ExpandingStd",
    "ExpandingMin",
    "ExpandingMax",
    "ExpandingQuantile",
    "ExponentiallyWeightedMean",
]

import abc
import copy
from typing import TYPE_CHECKING, Callable, Optional, Sequence

import numpy as np

if TYPE_CHECKING:
    from ._lib.grouped_array import _GroupedArrayFloat32, _GroupedArrayFloat64

    GroupedArray = _GroupedArrayFloat32 | _GroupedArrayFloat64


class _BaseLagTransform(abc.ABC):
    stats_: np.ndarray

    @abc.abstractmethod
    def transform(self, ga: "GroupedArray") -> np.ndarray:
        """Apply the transformation by group.

        Args:
            ga (GroupedArray): Array with the grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        ...

    @abc.abstractmethod
    def update(self, ga: "GroupedArray") -> np.ndarray:
        """Compute the most recent value of the transformation for each group.

        Args:
            ga (GroupedArray): Array with the grouped data.

        Returns:
            np.ndarray: Array with the updates for each group."""
        ...

    def take(self, _idxs: np.ndarray) -> "_BaseLagTransform":
        return self

    @staticmethod
    def stack(transforms: Sequence["_BaseLagTransform"]) -> "_BaseLagTransform":
        first_tfm = transforms[0]
        if not hasattr(first_tfm, "stats_"):
            # transform doesn't save state, we can return any of them
            return first_tfm
        out = copy.deepcopy(first_tfm)
        if first_tfm.stats_.ndim == 1:
            concat_fn = np.hstack
        else:
            concat_fn = np.vstack
        out.stats_ = concat_fn([tfm.stats_ for tfm in transforms])
        return out


class Lag(_BaseLagTransform):
    """Simple lag operator

    Args:
        lag (int): Number of periods to offset"""

    def __init__(self, lag: int):
        self.lag = lag

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        return ga._lag(self.lag)

    def update(self, ga: "GroupedArray") -> np.ndarray:
        return ga._index_from_end(self.lag - 1)


_rolling_base_docstring = """Rolling {stat_name}

    Args:
        lag (int): Number of periods to offset by before applying the transformation.
        window_size (int): Length of the rolling window.
        min_samples (int, optional): Minimum number of samples required to compute the statistic.
            If None, defaults to window_size."""


class _RollingBase(_BaseLagTransform):
    stat_name: str
    lag: int
    window_size: int
    min_samples: int

    def __init__(self, lag: int, window_size: int, min_samples: Optional[int] = None):
        self.lag = lag
        if min_samples is None:
            min_samples = window_size
        if min_samples > window_size:
            min_samples = window_size
        self.window_size = window_size
        self.min_samples = min_samples

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        return getattr(ga, f"_rolling_{self.stat_name}")(
            self.lag, self.window_size, self.min_samples
        )

    def update(self, ga: "GroupedArray") -> np.ndarray:
        return getattr(ga, f"_rolling_{self.stat_name}_update")(
            self.lag - 1, self.window_size, self.min_samples
        )


class RollingMean(_RollingBase):
    stat_name = "mean"


RollingMean.__doc__ = _rolling_base_docstring.format(stat_name="Mean")


class RollingStd(_RollingBase):
    stat_name = "std"


RollingStd.__doc__ = _rolling_base_docstring.format(stat_name="Standard Deviation")


class RollingMin(_RollingBase):
    stat_name = "min"


RollingMin.__doc__ = _rolling_base_docstring.format(stat_name="Minimum")


class RollingMax(_RollingBase):
    stat_name = "max"


RollingMax.__doc__ = _rolling_base_docstring.format(stat_name="Maximum")


class RollingQuantile(_RollingBase):
    """Rolling quantile

    Args:
        lag (int): Number of periods to offset by before applying the transformation
        p (float): Quantile to compute
        window_size (int): Length of the rolling window
        min_samples (int, optional): Minimum number of samples required to compute the statistic.
            If None, defaults to window_size."""

    def __init__(
        self, lag: int, p: float, window_size: int, min_samples: Optional[int] = None
    ):
        super().__init__(lag=lag, window_size=window_size, min_samples=min_samples)
        self.p = p

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        return ga._rolling_quantile(
            self.lag, self.p, self.window_size, self.min_samples
        )

    def update(self, ga: "GroupedArray") -> np.ndarray:
        return ga._rolling_quantile_update(
            self.lag - 1, self.p, self.window_size, self.min_samples
        )


_seasonal_rolling_docstring = """Seasonal rolling {stat_name}

    Args:
        lag (int): Number of periods to offset by before applying the transformation
        season_length (int): Length of the seasonal period, e.g. 7 for weekly data
        window_size (int): Length of the rolling window
        min_samples (int, optional): Minimum number of samples required to compute the statistic.
            If None, defaults to window_size."""


class _SeasonalRollingBase(_RollingBase):
    season_length: int

    def __init__(
        self,
        lag: int,
        season_length: int,
        window_size: int,
        min_samples: Optional[int] = None,
    ):
        super().__init__(lag, window_size, min_samples)
        self.season_length = season_length

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        return getattr(ga, f"_seasonal_rolling_{self.stat_name}")(
            self.lag,
            self.season_length,
            self.window_size,
            self.min_samples,
        )

    def update(self, ga: "GroupedArray") -> np.ndarray:
        return getattr(ga, f"_seasonal_rolling_{self.stat_name}_update")(
            self.lag - 1,
            self.season_length,
            self.window_size,
            self.min_samples,
        )


class SeasonalRollingMean(_SeasonalRollingBase):
    stat_name = "mean"


SeasonalRollingMean.__doc__ = _seasonal_rolling_docstring.format(stat_name="Mean")


class SeasonalRollingStd(_SeasonalRollingBase):
    stat_name = "std"


SeasonalRollingStd.__doc__ = _seasonal_rolling_docstring.format(
    stat_name="Standard Deviation"
)


class SeasonalRollingMin(_SeasonalRollingBase):
    stat_name = "min"


SeasonalRollingMin.__doc__ = _seasonal_rolling_docstring.format(stat_name="Minimum")


class SeasonalRollingMax(_SeasonalRollingBase):
    stat_name = "max"


SeasonalRollingMax.__doc__ = _seasonal_rolling_docstring.format(stat_name="Maximum")


class SeasonalRollingQuantile(_SeasonalRollingBase):
    """Seasonal rolling statistic

    Args:
        lag (int): Number of periods to offset by before applying the transformation
        p (float): Quantile to compute
        season_length (int): Length of the seasonal period, e.g. 7 for weekly data
        window_size (int): Length of the rolling window
        min_samples (int, optional): Minimum number of samples required to compute the statistic.
            If None, defaults to window_size."""

    def __init__(
        self,
        lag: int,
        p: float,
        season_length: int,
        window_size: int,
        min_samples: Optional[int] = None,
    ):
        super().__init__(
            lag=lag,
            season_length=season_length,
            window_size=window_size,
            min_samples=min_samples,
        )
        self.p = p

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        return ga._seasonal_rolling_quantile(
            self.lag,
            self.p,
            self.season_length,
            self.window_size,
            self.min_samples,
        )

    def update(self, ga: "GroupedArray") -> np.ndarray:
        return ga._seasonal_rolling_quantile_update(
            self.lag - 1,
            self.p,
            self.season_length,
            self.window_size,
            self.min_samples,
        )


_expanding_docstring = """Expanding {stat_name}

    Args:
        lag (int): Number of periods to offset by before applying the transformation"""


class _ExpandingBase(_BaseLagTransform):
    stats_: np.ndarray

    def __init__(self, lag: int):
        self.lag = lag

    def take(self, idxs: np.ndarray) -> "_ExpandingBase":
        out = self.__class__(self.lag)
        out.stats_ = self.stats_[idxs].copy()
        return out


class ExpandingMean(_ExpandingBase):
    def transform(self, ga: "GroupedArray") -> np.ndarray:
        out, n = ga._expanding_mean(self.lag)
        cumsum = n * out[ga.indptr[1:] - 1]
        self.stats_ = np.hstack([n[:, None], cumsum[:, None]])
        return out

    def update(self, ga: "GroupedArray") -> np.ndarray:
        self.stats_[:, 0] += 1.0
        self.stats_[:, 1] += ga._index_from_end(self.lag - 1)
        return self.stats_[:, 1] / self.stats_[:, 0]


ExpandingMean.__doc__ = _expanding_docstring.format(stat_name="Mean")


class ExpandingStd(_ExpandingBase):
    def transform(self, ga: "GroupedArray") -> np.ndarray:
        out, self.stats_ = ga._expanding_std(self.lag)
        return out

    def update(self, ga: "GroupedArray") -> np.ndarray:
        x = ga._index_from_end(self.lag - 1)
        self.stats_[:, 0] += 1.0
        n = self.stats_[:, 0]
        prev_avg = self.stats_[:, 1].copy()
        self.stats_[:, 1] = prev_avg + (x - prev_avg) / n
        self.stats_[:, 2] += (x - prev_avg) * (x - self.stats_[:, 1])
        self.stats_[:, 2] = np.maximum(self.stats_[:, 2], 0.0)
        return np.sqrt(self.stats_[:, 2] / (n - 1))


ExpandingStd.__doc__ = _expanding_docstring.format(stat_name="Standard Deviation")


class _ExpandingComp(_ExpandingBase):
    stat: str
    _comp_fn: Callable

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        out = getattr(ga, f"_expanding_{self.stat}")(self.lag)
        self.stats_ = out[ga.indptr[1:] - 1]
        return out

    def update(self, ga: "GroupedArray") -> np.ndarray:
        self.stats_ = self._comp_fn(self.stats_, ga._index_from_end(self.lag - 1))
        return self.stats_


class ExpandingMin(_ExpandingComp):
    stat = "min"
    _comp_fn = np.minimum


ExpandingMin.__doc__ = _expanding_docstring.format(stat_name="Minimum")


class ExpandingMax(_ExpandingComp):
    stat = "max"
    _comp_fn = np.maximum


ExpandingMax.__doc__ = _expanding_docstring.format(stat_name="Maximum")


class ExpandingQuantile(_BaseLagTransform):
    """Expanding quantile

    Args:
        lag (int):
            Number of periods to offset by before applying the transformation
        p (float):
            Quantile to compute"""

    def __init__(self, lag: int, p: float):
        self.lag = lag
        self.p = p

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        return ga._expanding_quantile(self.lag, self.p)

    def update(self, ga: "GroupedArray") -> np.ndarray:
        return ga._expanding_quantile_update(self.lag - 1, self.p)


class ExponentiallyWeightedMean(_BaseLagTransform):
    """Exponentially weighted mean

    Args:
        lag (int): Number of periods to offset by before applying the transformation
        alpha (float): Smoothing factor"""

    def __init__(self, lag: int, alpha: float):
        self.lag = lag
        self.alpha = alpha

    def transform(self, ga: "GroupedArray") -> np.ndarray:
        out = ga._exponentially_weighted_mean(self.lag, self.alpha)
        self.stats_ = out[ga.indptr[1:] - 1]
        return out

    def update(self, ga: "GroupedArray") -> np.ndarray:
        x = ga._index_from_end(self.lag - 1)
        self.stats_ = self.alpha * x + (1 - self.alpha) * self.stats_
        return self.stats_

    def take(self, idxs: np.ndarray) -> "ExponentiallyWeightedMean":
        out = copy.deepcopy(self)
        out.stats_ = out.stats_[idxs].copy()
        return out



__all__ = [
    "rolling_mean",
    "rolling_std",
    "rolling_min",
    "rolling_max",
    "rolling_quantile",
    "seasonal_rolling_mean",
    "seasonal_rolling_std",
    "seasonal_rolling_min",
    "seasonal_rolling_max",
    "seasonal_rolling_quantile",
]

from typing import Callable, Optional

import numpy as np

from ._lib import rolling as _rolling


def _rolling_stat(
    x: np.ndarray,
    stat: str,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    if min_samples is None:
        min_samples = window_size
    return getattr(_rolling, f"rolling_{stat}")(x, window_size, min_samples)


def _seasonal_rolling_stat(
    x: np.ndarray,
    stat: str,
    season_length: int,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    if min_samples is None:
        min_samples = window_size
    return getattr(_rolling, f"seasonal_rolling_{stat}")(
        x, season_length, window_size, min_samples
    )


def _rolling_docstring(*args, **kwargs) -> Callable:
    base_docstring = """Compute the {} of the input array.

    Args:
        x (np.ndarray): Input array.
        window_size (int): The size of the rolling window.
        min_samples (int, optional): The minimum number of samples required to compute the statistic.
            If None, it is set to `window_size`.

    Returns:
        np.ndarray: Array with the rolling statistic
    """

    def docstring_decorator(function: Callable):
        function.__doc__ = base_docstring.format(function.__name__)
        return function

    return docstring_decorator(*args, **kwargs)


def _seasonal_rolling_docstring(*args, **kwargs) -> Callable:
    base_docstring = """Compute the {} of the input array

    Args:
        x (np.ndarray): Input array.
        season_length (int): The length of the seasonal period.
        window_size (int): The size of the rolling window.
        min_samples (int, optional): The minimum number of samples required to compute the statistic.
            If None, it is set to `window_size`.

    Returns:
        np.ndarray: Array with the seasonal rolling statistic
    """

    def docstring_decorator(function: Callable):
        function.__doc__ = base_docstring.format(function.__name__)
        return function

    return docstring_decorator(*args, **kwargs)


@_rolling_docstring
def rolling_mean(
    x: np.ndarray, window_size: int, min_samples: Optional[int] = None
) -> np.ndarray:
    return _rolling_stat(x, "mean", window_size, min_samples)


@_rolling_docstring
def rolling_std(
    x: np.ndarray, window_size: int, min_samples: Optional[int] = None
) -> np.ndarray:
    return _rolling_stat(x, "std", window_size, min_samples)


@_rolling_docstring
def rolling_min(
    x: np.ndarray, window_size: int, min_samples: Optional[int] = None
) -> np.ndarray:
    return _rolling_stat(x, "min", window_size, min_samples)


@_rolling_docstring
def rolling_max(
    x: np.ndarray, window_size: int, min_samples: Optional[int] = None
) -> np.ndarray:
    return _rolling_stat(x, "max", window_size, min_samples)


def rolling_quantile(
    x: np.ndarray, p: float, window_size: int, min_samples: Optional[int] = None
) -> np.ndarray:
    """Compute the rolling_quantile of the input array.

    Args:
        x (np.ndarray): Input array.
        q (float): Quantile to compute.
        window_size (int): The size of the rolling window.
        min_samples (int, optional): The minimum number of samples required to compute the statistic.
            If None, it is set to `window_size`.

    Returns:
        np.ndarray: Array with rolling statistic
    """
    if min_samples is None:
        min_samples = window_size
    return _rolling.rolling_quantile(x, window_size, min_samples, p)


@_seasonal_rolling_docstring
def seasonal_rolling_mean(
    x: np.ndarray,
    season_length: int,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    return _seasonal_rolling_stat(x, "mean", season_length, window_size, min_samples)


@_seasonal_rolling_docstring
def seasonal_rolling_std(
    x: np.ndarray,
    season_length: int,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    return _seasonal_rolling_stat(x, "std", season_length, window_size, min_samples)


@_seasonal_rolling_docstring
def seasonal_rolling_min(
    x: np.ndarray,
    season_length: int,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    return _seasonal_rolling_stat(x, "min", season_length, window_size, min_samples)


@_seasonal_rolling_docstring
def seasonal_rolling_max(
    x: np.ndarray,
    season_length: int,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    return _seasonal_rolling_stat(x, "max", season_length, window_size, min_samples)


def seasonal_rolling_quantile(
    x: np.ndarray,
    p: float,
    season_length: int,
    window_size: int,
    min_samples: Optional[int] = None,
) -> np.ndarray:
    """Compute the seasonal_rolling_quantile of the input array.

    Args:
        x (np.ndarray): Input array.
        q (float): Quantile to compute.
        season_length (int): The length of the seasonal period.
        window_size (int): The size of the rolling window.
        min_samples (int, optional): The minimum number of samples required to compute the statistic.
            If None, it is set to `window_size`.

    Returns:
        np.ndarray: Array with rolling statistic
    """
    if min_samples is None:
        min_samples = window_size
    return _rolling.seasonal_rolling_quantile(
        x, season_length, window_size, min_samples, p
    )



__all__ = [
    "AutoDifferences",
    "AutoSeasonalDifferences",
    "AutoSeasonalityAndDifferences",
    "Difference",
    "LocalBoxCoxScaler",
    "LocalMinMaxScaler",
    "LocalRobustScaler",
    "LocalStandardScaler",
    "boxcox",
    "boxcox_lambda",
    "inv_boxcox",
]

import copy
from typing import TYPE_CHECKING, List, Optional, Sequence, TypeVar, Union

import numpy as np

from ._lib import scalers as _scalers
from ._lib.grouped_array import GroupedArray
from .utils import _diffs_to_indptr, _indptr_dtype

if TYPE_CHECKING:
    from ._lib.grouped_array import _GroupedArrayFloat32, _GroupedArrayFloat64

    GroupedArrayT = Union["_GroupedArrayFloat32", "_GroupedArrayFloat64"]


T = TypeVar("T", "_GroupedArrayFloat32", "_GroupedArrayFloat64")


def _boxcox_lambda_checks(
    season_length: Optional[int],
    lower: float,
    upper: float,
    method: str,
) -> None:
    available_methods = ("guerrero", "loglik")
    if method not in available_methods:
        raise NotImplementedError(
            f"Method {method} not implemented. Available methods are {available_methods}"
        )
    if lower >= upper:
        raise ValueError("lower must be less than upper")
    if method == "guerrero" and season_length is None:
        raise ValueError("season_length is required when method='guerrero'")


def boxcox_lambda(
    x: np.ndarray,
    method: str,
    season_length: Optional[int] = None,
    lower: float = -0.9,
    upper: float = 2.0,
) -> float:
    """Find optimum lambda for the Box-Cox transformation

    Args:
        x (np.ndarray): Array with data to transform.
        method (str): Method to use. Valid options are 'guerrero' and 'loglik'.
            'guerrero' minimizes the coefficient of variation for subseries of `x` and supports negative values.
            'loglik' maximizes the log-likelihood function.
        season_length (int, optional): Length of the seasonal period.
            Only required if method='guerrero'.
        lower (float): Lower bound for the lambda.
        upper (float): Upper bound for the lambda.

    Returns:
        float: Optimum lambda."""
    _boxcox_lambda_checks(
        method=method,
        season_length=season_length,
        lower=lower,
        upper=upper,
    )
    if method == "loglik" and any(x <= 0):
        raise ValueError("All values in x must be positive when method='loglik'")
    if method == "guerrero":
        assert season_length is not None
        res = _scalers.boxcox_lambda_guerrero(x, season_length, lower, upper)
    else:
        res = _scalers.boxcox_lambda_loglik(x, lower, upper)
    return res


def boxcox(x: np.ndarray, lmbda: float) -> np.ndarray:
    """Apply the Box-Cox transformation

    Args:
        x (np.ndarray): Array with data to transform.
        lmbda (float): Lambda value to use.

    Returns:
        np.ndarray: Array with the transformed data."""
    return _scalers.boxcox(x, lmbda)


def inv_boxcox(x: np.ndarray, lmbda: float) -> np.ndarray:
    """Invert the Box-Cox transformation

    Args:
        x (np.ndarray): Array with data to transform.
        lmbda (float): Lambda value to use.

    Returns:
        np.ndarray: Array with the inverted transformation."""
    return _scalers.inv_boxcox(x, lmbda)


def _validate_scaler_size(stats: np.ndarray, ga: "GroupedArrayT") -> None:
    if stats.shape[0] != len(ga):
        raise ValueError(
            f"Number of groups in the scaler ({stats.shape[0]:,}) "
            f"doesn't match the ones in data ({len(ga):,})."
        )


class _BaseLocalScaler:
    _scaler_type: str

    def fit(self, ga: "GroupedArrayT") -> "_BaseLocalScaler":
        """Compute the statistics for each group.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            self: The fitted scaler object."""
        self.stats_ = getattr(ga, f"_{self._scaler_type}_stats")()
        return self

    def transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Use the computed statistics to apply the transformation.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        _validate_scaler_size(self.stats_, ga)
        return ga._scaler_transform(self.stats_)

    def fit_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """ "Compute the statistics for each group and apply the transformation.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        return self.fit(ga).transform(ga)

    def inverse_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Use the computed statistics to invert the transformation.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the inverted transformation."""
        _validate_scaler_size(self.stats_, ga)
        return ga._scaler_inverse_transform(self.stats_)

    def take(self, idxs: np.ndarray) -> "_BaseLocalScaler":
        out = copy.deepcopy(self)
        out.stats_ = self.stats_[idxs].copy()
        return out

    @staticmethod
    def stack(scalers: Sequence["_BaseLocalScaler"]) -> "_BaseLocalScaler":
        out = copy.deepcopy(scalers[0])
        out.stats_ = np.vstack([sc.stats_ for sc in scalers])
        return out


class LocalMinMaxScaler(_BaseLocalScaler):
    """Scale each group to the [0, 1] interval"""

    _scaler_type = "minmax"


class LocalStandardScaler(_BaseLocalScaler):
    """Scale each group to have zero mean and unit variance"""

    _scaler_type = "standard"


class LocalRobustScaler(_BaseLocalScaler):
    """Scale each group using robust statistics

    Args:
        scale (str): Type of robust scaling to use. Valid options are 'iqr' and 'mad'.
            If 'iqr' will use the inter quartile range as the scale.
            If 'mad' will use median absolute deviation as the scale."""

    def __init__(self, scale: str):
        if scale == "iqr":
            self._scaler_type = "robust_iqr"
        else:
            self._scaler_type = "robust_mad"


class LocalBoxCoxScaler(_BaseLocalScaler):
    """Find the optimum lambda for the Box-Cox transformation by group and apply it

    Args:
        season_length (int, optional): Length of the seasonal period.
            Only required if method='guerrero'.
        lower (float): Lower bound for the lambda.
        upper (float): Upper bound for the lambda.
        method (str): Method to use. Valid options are 'guerrero' and 'loglik'.
            'guerrero' minimizes the coefficient of variation for subseries of `x` and supports negative values.
            'loglik' maximizes the log-likelihood function."""

    def __init__(
        self,
        method: str,
        season_length: Optional[int] = None,
        lower: float = -0.9,
        upper: float = 2.0,
    ):
        _boxcox_lambda_checks(season_length, lower, upper, method)
        self.method = method
        self.season_length = season_length
        self.lower = lower
        self.upper = upper

    def fit(self, ga: "GroupedArrayT") -> "_BaseLocalScaler":
        """Compute the statistics for each group.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            self: The fitted scaler object."""
        if self.method == "loglik" and any(ga.data < 0):
            raise ValueError("All values in data must be positive for method='loglik'")
        if self.method == "loglik":
            self.stats_ = ga._boxcox_loglik(
                self.lower,
                self.upper,
            )
        else:
            self.stats_ = ga._boxcox_guerrero(
                self.season_length,
                self.lower,
                self.upper,
            )
        return self

    def transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Use the computed lambdas to apply the transformation.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        _validate_scaler_size(self.stats_, ga)
        return ga._boxcox(self.stats_)

    def inverse_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Use the computed lambdas to invert the transformation.

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the inverted transformation."""
        _validate_scaler_size(self.stats_, ga)
        return ga._inv_boxcox(self.stats_)


class Difference:
    """Subtract a lag to each group

    Args:
        d (int): Lag to subtract."""

    def __init__(self, d: int):
        self.d = d

    def fit_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Apply the transformation

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        self.tails_ = ga._tail(self.d)
        return ga._diff(self.d)

    def inverse_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Invert the transformation

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the inverted transformation."""
        return ga._inv_diff(self.d, self.tails_)

    def update(self, ga: "GroupedArrayT") -> np.ndarray:
        """Update the last observations from each serie

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the updated data."""
        tails_indptr = np.arange(
            0, ga.indptr.size * self.d, self.d, dtype=_indptr_dtype
        )
        if self.tails_.size != tails_indptr[-1]:
            raise ValueError("Number of tails doesn't match the number of groups")
        tails_ga = GroupedArray(self.tails_, tails_indptr, num_threads=ga.num_threads)
        combined = tails_ga._append(ga)
        self.tails_ = combined._tail(self.d)
        combined_transformed = combined._diff(self.d)
        transformed = combined._with_data(combined_transformed)._tails(ga.indptr)
        return transformed

    def take(self, idxs: np.ndarray) -> "Difference":
        out = Difference(self.d)
        out.tails_ = self.tails_.reshape(-1, self.d)[idxs].ravel()
        return out

    @staticmethod
    def stack(scalers: Sequence["Difference"]) -> "Difference":
        out = Difference(scalers[0].d)
        out.tails_ = np.hstack([sc.tails_ for sc in scalers])
        return out


class AutoDifferences:
    """Find and apply the optimal number of differences to each group.

    Args:
        max_diffs (int): Maximum number of differences to apply."""

    def __init__(self, max_diffs: int):
        if not isinstance(max_diffs, int) or max_diffs <= 0:
            raise ValueError("max_diff must be a positive integer")
        self.max_diffs = max_diffs

    def _transform(self, ga: "GroupedArrayT", season_length: int) -> np.ndarray:
        max_d = int(self.diffs_.max())
        transformed = ga.data.copy()
        self.tails_: List[np.ndarray] = []
        for i in range(max_d):
            ga = ga._with_data(transformed)
            mask = self.diffs_ > i
            ds = np.where(mask, _indptr_dtype(season_length), _indptr_dtype(0))
            tails_indptr = _diffs_to_indptr(ds)
            self.tails_.append(ga._tails(tails_indptr))
            transformed = ga._diffs(ds)
        return transformed

    def fit_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Compute and apply the optimal number of differences for each group

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        self.diffs_ = ga._num_diffs(self.max_diffs)
        return self._transform(ga, season_length=1)

    def _inverse_transform(self, ga: "GroupedArrayT", season_length: int) -> np.ndarray:
        _validate_scaler_size(self.diffs_, ga)
        transformed = ga.data.copy()
        n_diffs = len(self.tails_)
        for i, tails in enumerate(self.tails_[::-1]):
            ga = ga._with_data(transformed)
            mask = self.diffs_ >= n_diffs - i
            ds = np.where(mask, _indptr_dtype(season_length), _indptr_dtype(0))
            transformed = ga._inv_diffs(ds, tails)
        return transformed

    def inverse_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Invert the differences

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the inverted transformation."""
        return self._inverse_transform(ga, 1)

    def _update(self, ga: "GroupedArrayT", season_length: int) -> np.ndarray:
        _validate_scaler_size(self.diffs_, ga)
        tail_indptr = np.arange(
            0, season_length * ga.indptr.size, season_length, dtype=_indptr_dtype
        )
        new_tails = []
        transformed = ga.data.copy()
        for i, tail in enumerate(self.tails_):
            ga = ga._with_data(transformed)
            tail_ga = GroupedArray(tail, tail_indptr, num_threads=ga.num_threads)
            combined = tail_ga._append(ga)
            new_tails.append(combined._tail(season_length))
            ds = np.where(
                self.diffs_ > i, _indptr_dtype(season_length), _indptr_dtype(0)
            )
            combined_transformed = combined._diffs(ds)
            transformed = combined._with_data(combined_transformed)._tails(ga.indptr)
        self.tails_ = new_tails
        return transformed

    def update(self, ga: "GroupedArrayT") -> np.ndarray:
        """Update the last observations from each serie

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the updated data."""
        return self._update(ga, season_length=1)

    def _take(self, idxs: np.ndarray, season_length: int) -> "AutoDifferences":
        out = copy.deepcopy(self)
        out.diffs_ = self.diffs_[idxs].copy()
        out.tails_ = []
        for i, tail in enumerate(self.tails_):
            ds = np.where(
                self.diffs_ > i, _indptr_dtype(season_length), _indptr_dtype(0)
            )
            tails_indptr = _diffs_to_indptr(ds)
            out.tails_.append(GroupedArray(tail, tails_indptr)._take(idxs))
        return out

    def take(self, idxs: np.ndarray) -> "AutoDifferences":
        return self._take(idxs, season_length=1)

    @staticmethod
    def stack(scalers: Sequence["AutoDifferences"]) -> "AutoDifferences":
        out = copy.deepcopy(scalers[0])
        out.diffs_ = np.hstack([sc.diffs_ for sc in scalers])
        out.tails_ = [
            np.hstack([sc.tails_[i] for sc in scalers])
            for i in range(len(scalers[0].tails_))
        ]
        return out


class AutoSeasonalDifferences(AutoDifferences):
    """Find and apply the optimal number of seasonal differences to each group.

    Args:
        season_length (int): Length of the seasonal period.
        max_diffs (int): Maximum number of differences to apply.
        n_seasons (int | None): Number of seasons to use to determine the number of differences. Defaults to 10.
            If `None` will use all samples, otherwise `season_length` * `n_seasons` samples will be used for the test.
            Smaller values will be faster but could be less accurate.
    """

    def __init__(
        self,
        season_length: int,
        max_diffs: int,
        n_seasons: Optional[int] = 10,
    ):
        self.season_length = season_length
        self.max_diffs = max_diffs
        if isinstance(n_seasons, int) and n_seasons < 2:
            raise ValueError("n_seasons must be at least 2")
        self.n_seasons = n_seasons

    def fit_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Compute and apply the optimal number of seasonal differences for each group

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        if self.n_seasons is None:
            tails_ga = ga
        else:
            n_samples = self.season_length * self.n_seasons
            tails_ga = GroupedArray(
                ga._tail(n_samples),
                np.arange(0, (len(ga) + 1) * n_samples, n_samples),
                num_threads=ga.num_threads,
            )
        self.diffs_ = tails_ga._num_seas_diffs(self.season_length, self.max_diffs)
        return self._transform(ga, self.season_length)

    def inverse_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Invert the seasonal differences

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the inverted transformation."""
        return self._inverse_transform(ga, self.season_length)

    def update(self, ga: "GroupedArrayT") -> np.ndarray:
        """Update the last observations from each serie

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the updated data."""
        return self._update(ga, self.season_length)

    def take(self, idxs: np.ndarray) -> "AutoDifferences":
        return self._take(idxs, self.season_length)


class AutoSeasonalityAndDifferences:
    """Find the length of the seasonal period and apply the optimal number of differences to each group.

    Args:
        max_season_length (int): Maximum length of the seasonal period.
        max_diffs (int): Maximum number of differences to apply.
        n_seasons (int | None): Number of seasons to use to determine the number of differences. Defaults to 10.
            If `None` will use all samples, otherwise `max_season_length` * `n_seasons` samples will be used for the test.
            Smaller values will be faster but could be less accurate.
    """

    def __init__(
        self, max_season_length: int, max_diffs: int, n_seasons: Optional[int] = 10
    ):
        self.max_season_length = max_season_length
        if not isinstance(max_diffs, int) or max_diffs < 1:
            raise ValueError("max_diff must be a positive integer")
        self.max_diffs = max_diffs
        if isinstance(n_seasons, int) and n_seasons < 2:
            raise ValueError("n_seasons must be at least 2")
        self.n_seasons = n_seasons

    def fit_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Compute the optimal length of the seasonal period and apply the optimal number of differences for each group

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the transformed data."""
        self.diffs_: List[np.ndarray] = []
        self.tails_: List[np.ndarray] = []
        transformed = ga.data.copy()
        for _ in range(self.max_diffs):
            ga = ga._with_data(transformed)
            if self.n_seasons is None:
                tails_ga = ga
            else:
                n_samples = self.max_season_length * self.n_seasons
                tails_ga = GroupedArray(
                    ga._tail(n_samples),
                    np.arange(0, (len(ga) + 1) * n_samples, n_samples),
                    num_threads=ga.num_threads,
                )
            periods = tails_ga._periods(self.max_season_length)
            diffs = periods * tails_ga._num_seas_diffs_periods(1, periods)
            diffs = diffs.astype(_indptr_dtype)
            self.diffs_.append(diffs)
            tails_indptr = _diffs_to_indptr(diffs)
            self.tails_.append(ga._tails(tails_indptr))
            transformed = ga._diffs(diffs)
        return transformed

    def inverse_transform(self, ga: "GroupedArrayT") -> np.ndarray:
        """Invert the seasonal differences

        Args:
            ga (GroupedArray): Array with grouped data.

        Returns:
            np.ndarray: Array with the inverted transformation."""
        _validate_scaler_size(self.diffs_[0], ga)
        transformed = ga.data.copy()
        for diffs, tails in zip(self.diffs_[::-1], self.tails_[::-1]):
            ga = ga._with_data(transformed)
            transformed = ga._inv_diffs(diffs, tails)
        return transformed

    def update(self, ga: "GroupedArrayT") -> np.ndarray:
        _validate_scaler_size(self.diffs_[0], ga)
        new_tails = []
        transformed = ga.data.copy()
        for i in range(self.max_diffs):
            ga = ga._with_data(transformed)
            tails_indptr = _diffs_to_indptr(self.diffs_[i])
            tails_ga = GroupedArray(
                self.tails_[i], tails_indptr, num_threads=ga.num_threads
            )
            combined = tails_ga._append(ga)
            new_tails.append(combined._tails(tails_indptr))
            combined_transformed = combined._diffs(self.diffs_[i])
            transformed = combined._with_data(combined_transformed)._tails(ga.indptr)
        self.tails_ = new_tails
        return transformed

    def take(self, idxs: np.ndarray) -> "AutoSeasonalityAndDifferences":
        out = AutoSeasonalityAndDifferences(
            self.max_season_length, self.max_diffs, self.n_seasons
        )
        out.diffs_ = [diffs[idxs].copy() for diffs in self.diffs_]
        out.tails_ = []
        for diffs, tails in zip(self.diffs_, self.tails_):
            tails_indptr = _diffs_to_indptr(diffs)
            tails_ga = GroupedArray(tails, tails_indptr)
            out.tails_.append(tails_ga._take(idxs))
        return out

    @staticmethod
    def stack(
        scalers: Sequence["AutoSeasonalityAndDifferences"],
    ) -> "AutoSeasonalityAndDifferences":
        out = AutoSeasonalityAndDifferences(
            scalers[0].max_season_length, scalers[0].max_diffs, scalers[0].n_seasons
        )
        out.diffs_ = [
            np.hstack([sc.diffs_[i] for sc in scalers]) for i in range(out.max_diffs)
        ]
        out.tails_ = [
            np.hstack([sc.tails_[i] for sc in scalers]) for i in range(out.max_diffs)
        ]
        return out
